<?php

namespace XF;

use Symfony\Component\Cache\Adapter;

use function call_user_func, in_array, intval, is_string;

class CacheFactory
{
	/**
	 * @var string
	 */
	protected $namespace = '';

	public function __construct(string $namespace = '')
	{
		$this->namespace = $namespace;
	}

	public function setNamespace(string $namespace): void
	{
		$this->namespace = $namespace;
	}

	public function getNamespace(): string
	{
		return $this->namespace;
	}

	/**
	 * @param string|\Closure $provider
	 * @param mixed[] $config
	 *
	 * @return Adapter\AdapterInterface|null
	 */
	public function create($provider, array $config = []): ?Adapter\AdapterInterface
	{
		return $this->instantiate($provider, $config);
	}

	/**
	 * @param string|\Closure $provider
	 * @param mixed[] $config
	 */
	protected function instantiate($provider, array $config): ?Adapter\AdapterInterface
	{
		// factory closure
		if ($provider instanceof \Closure)
		{
			$cache = $provider($config);
			if (!($cache instanceof Adapter\AdapterInterface))
			{
				$this->logDeprecatedCacheProvider('Closure');
				return null;
			}

			return $cache;
		}

		// class\name or class\name::method
		if (is_string($provider) && strpos($provider, '\\') !== false)
		{
			$parts = explode('::', $provider);
			if (isset($parts[1]))
			{
				// class::method - assume this is the factory method
				$cache = call_user_func($parts, $config);
			}
			else
			{
				// assume this is the provider itself
				$cache = new $provider($config);
			}

			if (!($cache instanceof Adapter\AdapterInterface))
			{
				$this->logDeprecatedCacheProvider($provider);
				return null;
			}

			return $cache;
		}

		// provider name only, which we'll map to a method
		if (is_string($provider) && $provider)
		{
			if (in_array($provider, ['Apc', 'Void', 'WinCache', 'XCache']))
			{
				$this->logDeprecatedCacheProvider($provider);
				return null;
			}

			$method = 'create' . $provider . 'Cache';
			if (!is_callable([$this, $method]))
			{
				throw new \InvalidArgumentException("Invalid cache provider '$provider'");
			}

			return $this->$method($config);
		}

		throw new \InvalidArgumentException("Invalid type of cache provider");
	}

	protected function logDeprecatedCacheProvider(string $provider): void
	{
		// log an error once per day per provider
		$registryKey = substr('depCache_' . md5($provider), 0, 25);
		\XF::runOnce($registryKey, function () use ($provider, $registryKey)
		{
			$registry = \XF::registry();
			$time = $registry->get($registryKey);
			if ($time !== null && $time >= \XF::$time - 86400)
			{
				return;
			}

			\XF::logError('Deprecated cache provider: ' . $provider);
			$registry->set($registryKey, \XF::$time);
		});
	}

	/**
	 * @param mixed[] $config
	 */
	protected function createApcuCache(array $config): Adapter\ApcuAdapter
	{
		return new Adapter\ApcuAdapter($this->namespace);
	}

	/**
	 * @param mixed[] $config
	 */
	protected function createFilesystemCache(array $config): Adapter\FilesystemAdapter
	{
		if (empty($config['directory']))
		{
			throw new \LogicException("Filesystem cache config must define a 'directory'");
		}

		$cache = new Adapter\FilesystemAdapter($this->namespace, 0, $config['directory']);

		$cleanFrequency = max(!empty($config['clean']) ? intval($config['clean']) : 1000, 2);
		if (rand(1, $cleanFrequency) == 1)
		{
			$cache->prune();
		}

		return $cache;
	}

	/**
	 * @param mixed[] $config
	 */
	protected function createMemcachedCache(array $config): Adapter\MemcachedAdapter
	{
		if (empty($config['servers']))
		{
			if (!empty($config['server']))
			{
				// legacy format
				$config['servers'] = [['host' => $config['server']]];
				unset($config['server']);
			}
			else
			{
				// default value
				$config['servers'] = [['host' => 'localhost']];
			}
		}

		$servers = [];
		foreach ($config['servers'] AS $server)
		{
			$server = 'memcached://' . $server['host'];

			$port = $server['port'] ?? null;
			if ($port !== null)
			{
				$server .= ':' . $port;
			}

			$weight = $server['weight'] ?? null;
			if ($weight !== null)
			{
				$server .= '?weight=' . $weight;
			}

			// TODO: auth support?

			$servers[] = $server;
		}

		$memcached = Adapter\MemcachedAdapter::createConnection($servers);

		if (!empty($config['custom']) && $config['custom'] instanceof \Closure)
		{
			$custom = $config['custom'];
			$custom($memcached);
		}

		return new Adapter\MemcachedAdapter($memcached, $this->namespace);
	}

	protected function createNullCache(array $config): Adapter\NullAdapter
	{
		return new Adapter\NullAdapter();
	}

	/**
	 * @param mixed[] $config
	 */
	protected function createRedisCache(array $config): Adapter\RedisAdapter
	{
		$config = array_replace([
			'persistent' => false,
			'persistent_id' => ''
		], $config);

		$dsn = 'redis://';

		$password = $config['password'] ?? null;
		if ($password !== null)
		{
			$dsn .= $password . '@';
		}

		$dsn .= $config['host'];

		$port = $config['port'] ?? null;
		if ($port !== null)
		{
			$dsn .= ':' . $port;
		}

		$database = $config['database'] ?? null;
		if ($database !== null)
		{
			$dsn .= '/' . $database;
		}

		$timeout = $config['timeout'] ?? null;
		if ($timeout !== null)
		{
			$dsn .= '?timeout=' . $timeout;
		}

		// TODO: persistence...?
		$options = [];

		$redis = Adapter\RedisAdapter::createConnection($dsn, $options);
		if (!empty($config['custom']) && $config['custom'] instanceof \Closure)
		{
			$custom = $config['custom'];
			$custom($redis);
		}

		return new Adapter\RedisAdapter($redis, $this->namespace);
	}
}