<?php

namespace XF\Cli\Command\Development;

use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use XF\Cli\Command\AbstractCommand;

class GenerateFinders extends AbstractCommand
{
	use RequiresDevModeTrait;

	protected function configure()
	{
		$this
			->setName('xf-dev:generate-finders')
			->setDescription('Generates skeleton finders from entities')
			->addArgument(
				'addon',
				InputOption::VALUE_REQUIRED,
				'Add-on ID to generate Finder classes for.'
			);
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$ds = \XF::$DS;
		$addOnId = $input->getArgument('addon');
		if ($addOnId === 'XF')
		{
			$searchDir = \XF::getSourceDirectory() . $ds . 'XF' . $ds . 'Entity';
		}
		else
		{
			$manager = \XF::app()->addOnManager();
			$addOn = $manager->getById($addOnId);
			if (!$addOn || !$addOn->isAvailable())
			{
				$output->writeln('Add-on could not be found.');
				return 1;
			}

			$addOnId = $addOn->getAddOnId();
			$searchDir = $manager->getAddOnPath($addOnId) . $ds . 'Entity';

			$addOnId = str_replace('/', '\\', $addOnId);
		}

		if (!file_exists($searchDir) || !is_dir($searchDir))
		{
			$output->writeln('<error>The selected add-on does not appear to have an Entity directory.</error>');
			return 1;
		}

		$iterator = new \RegexIterator(
			\XF\Util\File::getRecursiveDirectoryIterator($searchDir, null, null),
			'/\.php$/'
		);

		/** @var \SplFileInfo $file */
		foreach ($iterator AS $name => $file)
		{
			$suffixClass = str_replace($searchDir, '', $file->getRealPath());
			$suffixClass = str_replace('.php', '', $suffixClass);
			$suffixClass = trim(str_replace(\XF::$DS, '\\', $suffixClass), '\\');
			$fqClass = sprintf('%s\Entity\%s', "\\$addOnId", $suffixClass);
			$namespace = sprintf('%s\Finder', $addOnId);
			$fqResultingClass = sprintf('%s\Finder\%s', "\\$addOnId", $suffixClass);

			if (!class_exists($fqClass))
			{
				continue;
			}

			$reflection = new \ReflectionClass($fqClass);
			if ($reflection->isAbstract() || $reflection->isInterface() || $reflection->isTrait())
			{
				continue;
			}

			$docPlaceholder = $this->getDocPlaceholder();

			$parts = explode('\\', $namespace);
			if ($parts[0] === 'XF')
			{
				$path = \XF::getSourceDirectory() . \XF::$DS . implode(\XF::$DS, $parts);
			}
			else
			{
				$path = \XF::getAddOnDirectory() . \XF::$DS . implode(\XF::$DS, $parts);
			}

			$fileName = $path . \XF::$DS . $suffixClass . '.php';

			if (!class_exists($fqResultingClass))
			{
				// Generating new finder

				$fileOutput = <<< FINDEROUT
<?php

namespace $namespace;

use XF\Mvc\Entity\Finder;

/** <XF:DOC_COMMENT> */
class $suffixClass extends Finder
{
}

FINDEROUT;
			}
			else
			{
				// Updating existing finder

				$finderReflection = new \ReflectionClass($fqResultingClass);

				$contents = file_get_contents($fileName);
				$existingComment = $finderReflection->getDocComment();

				if (!$existingComment)
				{
					$search = 'class ' . $suffixClass . ' extends ';
					$replace = "$docPlaceholder\n$search";
					$fileOutput = str_replace($search, $replace, $contents);
				}
				else
				{
					$fileOutput = str_replace($existingComment, $docPlaceholder, $contents);
				}
			}

			$newComment = <<< COMMENTOUT
/**
 * @method \\XF\Mvc\Entity\AbstractCollection<$fqClass> fetch(\$limit = null, \$offset = null)
 * @method $fqClass|null fetchOne(\$limit = null, \$offset = null)
 * @extends \\XF\Mvc\Entity\Finder<$fqClass>
 */
COMMENTOUT;

			$fileOutput = str_replace($docPlaceholder, $newComment, $fileOutput);
			$output->writeln("Writing $fileName...");

			if (!is_writable(dirname($fileName)))
			{
				$output->writeln("File for $fileName could not be written to. Check directories exist and permissions.");
				return 5;
			}

			file_put_contents($fileName, $fileOutput);
			$output->writeln("Written $fileName successfully.");
		}

		$output->writeln("Done!");
		return 0;
	}

	protected function getDocPlaceholder(): string
	{
		return '/** <XF:DOC_COMMENT> */';
	}
}
