<?php

namespace XF\FeaturedContent;

use XF\Entity\ContainableInterface;
use XF\Entity\DatableInterface;
use XF\Entity\FeaturedContent;
use XF\Entity\User;
use XF\Mvc\Entity\AbstractCollection;
use XF\Mvc\Entity\Entity;

use function get_class;

abstract class AbstractHandler
{
	/**
	 * @var int
	 */
	const SNIPPET_LENGTH = 500;

	/**
	 * @var string
	 */
	protected $contentType;

	public function __construct(string $contentType)
	{
		$this->contentType = $contentType;
	}

	public function getContentType(): string
	{
		return $this->contentType;
	}

	/**
	 * Determines whether or not the given content may be viewed.
	 */
	public function canViewContent(Entity $content, &$error = null): bool
	{
		if (method_exists($content, 'canView'))
		{
			return $content->canView($error);
		}

		throw new \LogicException(
			'Override ' . get_class($this) . '::canViewContent'
		);
	}

	/**
	 * Determines the container ID of the given content.
	 */
	public function getContentContainerId(Entity $content): int
	{
		if (!($content instanceof ContainableInterface))
		{
			throw new \LogicException(
				'Could not determine content container ID; please override'
			);
		}

		return $content->getContentContainerId();
	}

	/**
	 * Determines the user ID of the user who created the given content.
	 */
	public function getContentUserId(Entity $content): int
	{
		if (isset($content->user_id))
		{
			return $content->user_id;
		}

		if (isset($content->User) && $content->User instanceof User)
		{
			return $content->User->user_id;
		}

		throw new \LogicException(
			'Override ' . get_class($this) . '::getContentUserId'
		);
	}

	/**
	 * Determines the username of the user who created the given content.
	 */
	public function getContentUsername(Entity $content): string
	{
		if (isset($content->username))
		{
			return $content->username;
		}

		if (isset($content->User) && $content->User instanceof User)
		{
			return $content->User->username;
		}

		throw new \LogicException(
			'Override ' . get_class($this) . '::getContentUsername'
		);
	}

	/**
	 * Determines the date the given content was created.
	 */
	public function getContentDate(Entity $content): int
	{
		if (!($content instanceof DatableInterface))
		{
			throw new \LogicException(
				'Could not determine content date; please override'
			);
		}

		return $content->getContentDate();
	}

	/**
	 * Determines whether the given content is visible.
	 */
	public function getContentVisibility(Entity $content): bool
	{
		if (method_exists($content, 'isVisible'))
		{
			return $content->isVisible();
		}

		throw new \LogicException(
			'Override ' . get_class($this) . '::getContentVisibility'
		);
	}

	/**
	 * Determines the title of the given content.
	 *
	 * This will be displayed on the feature when a custom title is not set.
	 */
	public function getContentTitle(Entity $content): string
	{
		if ($content instanceof \XF\Entity\LinkableInterface)
		{
			$title = $content->getContentTitle('featured_content');

			if ($title instanceof \XF\Phrase)
			{
				$title = $title->render('raw');
			}

			return $title;
		}

		throw new \LogicException(
			'Implement XF\Entity\LinkableInterface for ' . get_class($content)
			. ' or override ' . get_class($this) . '::getContentTitle'
		);
	}

	/**
	 * Determines the image of the given content.
	 *
	 * This will be displayed on the feature when a custom image is not set.
	 * Note that the size code is not currently utilized.
	 */
	abstract public function getContentImage(
		Entity $content,
		?string $sizeCode = null
	): ?string;

	/**
	 * Determines the snippet of the given content.
	 *
	 * This will be displayed on the feature when a custom snippet is not set.
	 */
	abstract public function getContentSnippet(Entity $content): string;

	protected function getSnippetFromString(?string $string): string
	{
		if (!$string)
		{
			return '';
		}

		$stringFormatter = \XF::app()->stringFormatter();
		return $stringFormatter->snippetString(
			$string,
			static::SNIPPET_LENGTH,
			['stripQuote' => true]
		);
	}

	public function getContentStructuredData(Entity $content): array
	{
		return [];
	}

	/**
	 * Determines the link of the given content.
	 */
	public function getContentLink(Entity $content): string
	{
		if ($content instanceof \XF\Entity\LinkableInterface)
		{
			return $content->getContentUrl();
		}

		throw new \LogicException(
			'Implement XF\Entity\LinkableInterface for ' . get_class($content)
			. ' or override ' . get_class($this) . '::getContentLink'
		);
	}

	/**
	 * Determines whether or not the given content should be automatically
	 * featured.
	 *
	 * This is called by the featurable behavior when the entity is first
	 * inserted.
	 */
	public function shouldAutoFeature(Entity $content): bool
	{
		return false;
	}

	/**
	 * A hook called when the given content is featured.
	 */
	public function onContentFeature(
		Entity $content,
		FeaturedContent $feature
	): void
	{
		$content->fastUpdate('featured', true);

		\XF::repository(\XF\Repository\Webhook::class)->queueWebhook(
			$content->getEntityContentType(), $content->getEntityId(), 'feature', $content
		);
	}

	/**
	 * A hook called when the given content is unfeatured.
	 */
	public function onContentUnfeature(
		Entity $content,
		FeaturedContent $feature
	): void
	{
		$content->fastUpdate('featured', false);

		\XF::repository(\XF\Repository\Webhook::class)->queueWebhook(
			$content->getEntityContentType(), $content->getEntityId(), 'unfeature', $content
		);
	}

	/**
	 * Renders the given feature for display on the featured content page.
	 *
	 * If the macro does not exist in the content-type template, it will
	 * fallback to the global macro.
	 */
	public function render(
		FeaturedContent $feature,
		string $macroName,
		int $snippetLength = 0
	): string
	{
		if (!$feature->Content)
		{
			return '';
		}

		$templateName = $this->getTemplateName();
		$macroArguments = $this->getMacroArguments(
			$macroName,
			$feature,
			$snippetLength
		);

		$templater = \XF::app()->templater();
		if ($templater->isKnownMacro($templateName, $macroName))
		{
			return $templater->renderMacro(
				$templateName,
				$macroName,
				$macroArguments
			);
		}

		return $templater->renderMacro(
			'featured_content_item',
			$macroName,
			$macroArguments
		);
	}

	/**
	 * Returns the template to use for rendering features for this content
	 * type.
	 */
	protected function getTemplateName(): string
	{
		$contentType = $this->contentType;
		return "public:featured_content_item_{$contentType}";
	}

	/**
	 * Returns the macro arguments to use for rendering the given feature macro.
	 */
	protected function getMacroArguments(
		string $macroName,
		FeaturedContent $feature,
		int $snippetLength = 0
	): array
	{
		$content = $feature->Content;

		return [
			'feature' => $feature,
			'content' => $content,
			'snippetLength' => $snippetLength,
		];
	}

	/**
	 * Fetches the content given a content ID or an array of content IDs.
	 *
	 * @param int|int[] $contentId
	 *
	 * @return \XF\Mvc\Entity\Entity|AbstractCollection|null
	 */
	public function getContent($contentId)
	{
		$content = \XF::app()->findByContentType(
			$this->contentType,
			$contentId,
			$this->getEntityWith()
		);
		if (!$content)
		{
			return $content;
		}

		if ($content instanceof AbstractCollection)
		{
			$isCollection = true;
		}
		else
		{
			$isCollection = false;
			$content = \XF::em()->getBasicCollection([$content]);
		}

		$imagelessContent = $content->filter(function(Entity $entity)
		{
			/** @var FeaturedContent $feature */
			$feature = $entity->Feature;
			return !$feature->image_date;
		});
		$this->addAttachmentsToContent($imagelessContent, $content);

		return $isCollection ? $content : $content->first();
	}

	/**
	 * Returns the relations to eager-load when fetching content for this
	 * content type.
	 */
	public function getEntityWith(): array
	{
		return [];
	}

	/**
	 * Adds attachments to a collection of content entities.
	 *
	 * This is typically only necessary when the default content image is
	 * derived from the content attachments.
	 *
	 * @param AbstractCollection $imagelessContent A collection of content entities with no custom feature image
	 * @param AbstractCollection $content A collection of all content entities
	 */
	protected function addAttachmentsToContent(
		AbstractCollection $imagelessContent,
		AbstractCollection $content
	): void
	{
	}

	/**
	 * A mechanism for hydrating attachments for a collection of content
	 * entities.
	 *
	 * This may be called within a handler's addAttachmentsToContent method.
	 */
	protected function addAttachmentsToContentInternal(
		AbstractCollection $content,
		string $countKey = 'attach_count',
		string $relationKey = 'Attachments'
	): void
	{
		/** @var \XF\Repository\Attachment $attachmentRepo */
		$attachmentRepo = \XF::repository('XF:Attachment');
		$attachmentRepo->addAttachmentsToContent(
			$content,
			$this->contentType,
			$countKey,
			$relationKey
		);
	}

	public static function getWebhookEvents(): array
	{
		return ['feature', 'unfeature'];
	}
}