<?php

namespace XF\Job;

use Minishlink\WebPush\Encryption;
use Minishlink\WebPush\MessageSentReport;
use Minishlink\WebPush\Subscription;
use Minishlink\WebPush\WebPush;

class PushSend extends AbstractJob
{
	public const GCM_URL = 'https://android.googleapis.com/gcm/send';

	const DELETE_ON_ERROR = true;

	protected $defaultData = [
		'payload' => null,
		'receiverUserId' => 0
	];

	public function run($maxRunTime): JobResult
	{
		if (!$this->data['payload'])
		{
			throw new \InvalidArgumentException('Cannot send push notification without a valid payload');
		}

		$receiver = $this->app->em()->find('XF:User', $this->data['receiverUserId']);
		if (!$receiver)
		{
			return $this->complete();
		}

		$payload = $this->data['payload'];
		$webPush = $this->getWebPushObject();
		$webPush->setDefaultOptions([
			'TTL' => 86400, // if undelivered after 1 day, expire the notification
		]);

		$subscriptions = $this->getUserPushRepository()->getUserSubscriptions($receiver);

		foreach ($subscriptions AS $subscription)
		{
			if (strpos($subscription['endpoint'], self::GCM_URL) === 0)
			{
				// if we somehow get a GCM URL into this, it will fail later, so we need to skip it here
				continue;
			}

			$authData = json_decode($subscription['data'], true);

			try
			{
				$webPush->setAutomaticPadding($this->getEndpointPadding($subscription['endpoint']));

				$subObj = Subscription::create([
					'endpoint' => $subscription['endpoint'],
					'publicKey' => $authData['key'],
					'authToken' => $authData['token'],
					'contentEncoding' => $authData['encoding']
				]);

				$webPush->queueNotification(
					$subObj,
					json_encode($payload)
				);
			}
			catch (\Exception $e)
			{
				// generally indicates that the payload is too big which at ~3000 bytes shouldn't happen for a typical alert...
				if (\XF::$debugMode)
				{
					\XF::logException($e);
				}
				continue;
			}
		}

		$results = $webPush->flush();
		$this->handleResults($results);

		return $this->complete();
	}

	/**
	 * @return WebPush
	 * @throws \ErrorException
	 */
	protected function getWebPushObject(): WebPush
	{
		$options = $this->app->options();

		$auth = [
			'VAPID' => [
					'subject' => $options->boardUrl
				] + $options->pushKeysVAPID
		];

		$httpOptions = $this->app->http()->getDefaultClientOptions();

		$config = $this->app->config();
		if ($config['http']['proxy'])
		{
			$httpOptions['proxy'] = $config['http']['proxy'];
		}

		return new WebPush($auth, [], 10, $httpOptions);
	}

	protected function getEndpointPadding($endpoint): int
	{
		if (strpos($endpoint, 'mozilla') !== false)
		{
			// firefox, at least on Android, has an issue with automatic padding which
			// is used to make encryption more secure, but makes encryption slower.
			// see: https://github.com/web-push-libs/web-push-php/issues/108
			// TODO: ideally won't need this forever if Mozilla ever fix this or if library works-around it
			return 0;
		}

		if (strpos($endpoint, '.ucweb.com') !== false)
		{
			// See https://xenforo.com/community/threads/158252/
			return 0;
		}

		return Encryption::MAX_COMPATIBILITY_PAYLOAD_LENGTH;
	}

	protected function handleResults($results): void
	{
		$db = $this->app->db();

		$pushRepo = $this->getUserPushRepository();

		/** @var MessageSentReport $report */
		foreach ($results AS $report)
		{
			$endpoint = (string)$report->getRequest()->getUri();
			$endpointHash = $pushRepo->getEndpointHash($endpoint);

			if ($report->isSuccess())
			{
				$db->update('xf_user_push_subscription', [
					'last_seen' => time()
				], 'endpoint_hash = ?', $endpointHash);
			}
			else
			{
				$response = $report->getResponse();
				$code = $response ? $response->getStatusCode() : null;

				switch ($code)
				{
					// 401 and 403 generally represent when something in the request doesn't match, likely
					// due to VAPID keys changing
					case 401:
					case 403:
					case 404:
					case 410:
						if (self::DELETE_ON_ERROR)
						{
							$db->delete('xf_user_push_subscription', 'endpoint_hash = ?', $endpointHash);
						}
						else
						{
							$errorCount = $db->fetchOne('SELECT error_count FROM xf_user_push_subscription WHERE endpoint_hash = ?', $endpointHash);
							$errorCount++;

							if ($errorCount >= 5)
							{
								$db->delete('xf_user_push_subscription', 'endpoint_hash = ?', $endpointHash);
							}
							else
							{
								$db->update('xf_user_push_subscription', [
									'error_count' => $errorCount
								], 'endpoint_hash = ?', $endpointHash);
							}
						}
						break;

					case 406:
						// not a server error but rate limiting - future pushes should work
					case 500:
					case 502:
					case 503:
					case 504:
						// these indicate server errors that are likely temporary - future pushes should work
						break;

					default:
						\XF::logError("Push notification failure: " . $report->getReason());
						break;
				}
			}
		}
	}

	public function getStatusMessage(): string
	{
		return \XF::phrase('sending_push_notifications...');
	}

	public function canCancel(): bool
	{
		return false;
	}

	public function canTriggerByChoice(): bool
	{
		return false;
	}

	protected function getUserPushRepository(): \XF\Repository\UserPush
	{
		return $this->app->repository('XF:UserPush');
	}
}