<?php

namespace XF\Payment;

use Psr\Http\Message\ResponseInterface;
use XF\Entity\PaymentProfile;
use XF\Entity\PurchaseRequest;
use XF\Mvc\Controller;
use XF\PrintableException;
use XF\Purchasable\Purchase;

use function in_array;

class PayPalRest extends AbstractProvider
{
	public function getTitle()
	{
		return 'PayPal';
	}

	public function renderConfig(PaymentProfile $profile)
	{
		$data = [
			'profile' => $profile,
			'events' => $this->getActionableEvents()
		];
		return \XF::app()->templater()->renderTemplate('admin:payment_profile_' . $this->providerId, $data);
	}

	public function getApiEndpoint()
	{
		if (\XF::config('enableLivePayments'))
		{
			return 'https://api-m.paypal.com/';
		}
		else
		{
			return 'https://api-m.sandbox.paypal.com/';
		}
	}

	public function verifyConfig(array &$options, &$errors = [])
	{
		return (bool)$this->getAccessToken($options['client_id'], $options['secret_key'], $errors);
	}

	protected function makePayPalRequest(string $method, string $endpoint, array $options = [], string $token = null, &$errors = []): array
	{
		$options = array_merge([
			'http_errors' => false,
			'headers' => [
				'Content-Type' => 'application/json'
			],
		], $options);

		if ($token !== null)
		{
			$options['headers']['Authorization'] = 'Bearer ' . $token;
		}

		$client = \XF::app()->http()->client();

		try
		{
			/** @var ResponseInterface $response */
			$response = $client->$method($this->getApiEndpoint() . $endpoint, $options);
			$responseBody = json_decode($response->getBody()->getContents(), true);

			if ($response->getStatusCode() > 204)
			{
				$errors[] = $responseBody['message'] ?? $responseBody['error_description'] ?? 'Unknown error';
			}

			return $responseBody;
		}
		catch (\GuzzleHttp\Exception\BadResponseException $e)
		{
			$errors[] = "PayPal API error ($endpoint): " . $e->getMessage();
		}

		return [];
	}

	protected $accessToken;

	protected function getAccessToken(string $clientId, string $secretKey, &$errors = []): ?string
	{
		if (!$this->accessToken)
		{
			$response = $this->makePayPalRequest('post', 'v1/oauth2/token', [
				'auth' => [$clientId, $secretKey],
				'form_params' => [
					'grant_type' => 'client_credentials'
				]
			], null, $errors);

			$this->accessToken = $response['access_token'] ?? null;
		}

		return $this->accessToken;
	}

	/**
	 * @throws PrintableException
	 */
	protected function assertAccessToken(string $clientId, string $secretKey): ?string
	{
		$accessToken = $this->getAccessToken($clientId, $secretKey, $errors);

		if (!$accessToken)
		{
			$errors = $errors ?? [];
			\XF::logError('Attempt to get PayPal access token failed: ' . reset($errors), true);
			throw new PrintableException(\XF::phrase('something_went_wrong_please_try_again'));
		}

		return $accessToken;
	}

	protected function getSubscriptionParams(string $planId, PurchaseRequest $purchaseRequest, Purchase $purchase)
	{
		return [
			'plan_id' => $planId,
			'quantity' => '1',
			'custom_id' => $purchaseRequest->request_key,
			'application_context' => [
				'return_url' => $purchase->returnUrl,
				'cancel_url' => $purchase->cancelUrl,
			]
		];
	}

	protected function getPaymentParams(PurchaseRequest $purchaseRequest, Purchase $purchase): array
	{
		return [
			'intent' => 'CAPTURE',
			'purchase_units' => [
				[
					'amount' => [
						'currency_code' => $purchase->currency,
						'value' => $purchase->cost
					],
					'description' => $purchase->title,
					'custom_id' => $purchaseRequest->request_key
				]
			],
			'payment_source' => [
				'paypal' => [
					'experience_context' => [
						'payment_method_preference' => 'IMMEDIATE_PAYMENT_REQUIRED',
						'user_action' => 'PAY_NOW',
						'return_url' => $purchase->returnUrl,
						'cancel_url' => $purchase->cancelUrl,
					]
				]
			]
		];
	}

	public function getPlanByProductId(Purchase $purchase, string $productId): array
	{
		$paymentProfile = $purchase->paymentProfile;
		$options = $paymentProfile->options;

		$accessToken = $this->assertAccessToken($options['client_id'], $options['secret_key']);

		$planDetails = $this->makePayPalRequest('get', 'v1/billing/plans', [
			'json' => [
				'product_id' => $productId,
				'page_size' => 1,
			]
		], $accessToken, $errors);

		if ($errors)
		{
			\XF::logError("Problem retrieving PayPal plans for product ($productId): " . reset($errors));
			throw new PrintableException(\XF::phrase('something_went_wrong_please_try_again'));
		}

		$plan = reset($planDetails['plans']);

		if (isset($plan['id']))
		{
			return $plan;
		}

		$params = [
			'product_id' => $productId,
			'name' => substr($purchase->title, 0, 127),
			'status' => 'ACTIVE',
			'billing_cycles' => [
				[
					'frequency' => [
						'interval_unit' => strtoupper($purchase->lengthUnit),
						'interval_count' => $purchase->lengthAmount,
					],
					'tenure_type' => 'REGULAR',
					'sequence' => 1,
					'total_cycles' => 0,
					'pricing_scheme' => [
						'fixed_price' => [
							'value' => $purchase->cost,
							'currency_code' => $purchase->currency,
						],
					],
				],
			],
			'payment_preferences' => [
				'payment_failure_threshold' => 2,
			],
		];

		$plan = $this->makePayPalRequest('post', 'v1/billing/plans', [
			'json' => $params
		], $accessToken, $errors);

		if ($errors)
		{
			\XF::logError("Problem creating PayPal plan for product ($productId): " . reset($errors));
			throw new PrintableException(\XF::phrase('something_went_wrong_please_try_again'));
		}

		return $plan;
	}

	/**
	 * @throws PrintableException
	 */
	protected function getProductById(Purchase $purchase, string $productId): array
	{
		$paymentProfile = $purchase->paymentProfile;
		$options = $paymentProfile->options;

		$accessToken = $this->assertAccessToken($options['client_id'], $options['secret_key']);
		$productDetails = $this->makePayPalRequest('get', "v1/catalogs/products/$productId", [], $accessToken, $errors);

		if ($errors)
		{
			\XF::logError("Problem getting PayPal product ($productId): " . reset($errors));
			throw new PrintableException(\XF::phrase('something_went_wrong_please_try_again'));
		}

		if (!isset($productDetails['id']))
		{
			$productDetails = $this->makePayPalRequest('post', 'v1/catalogs/products', [
				'json' => [
					'id' => $productId,
					'name' => $purchase->title,
					'type' => 'DIGITAL',
					'category' => 'MISCELLANEOUS_GENERAL_SERVICES',
				]
			], $accessToken, $errors);

			if ($errors)
			{
				\XF::logError("Problem creating PayPal product ($productId): " . reset($errors));
				throw new PrintableException(\XF::phrase('something_went_wrong_please_try_again'));
			}
		}

		return $productDetails;
	}

	public function initiatePayment(Controller $controller, PurchaseRequest $purchaseRequest, Purchase $purchase)
	{
		$paymentProfile = $purchase->paymentProfile;
		$options = $paymentProfile->options;

		$accessToken = $this->assertAccessToken($options['client_id'], $options['secret_key']);

		if ($purchase->recurring)
		{
			$productId = "{$purchase->purchasableTypeId}_{$purchase->purchasableId}";

			// note: creates product if it doesn't exist so must be called first
			$product = $this->getProductById($purchase, $productId);
			$plan = $this->getPlanByProductId($purchase, $product['id']);

			$setupResponse = $this->makePayPalRequest('post', 'v1/billing/subscriptions', [
				'json' => $this->getSubscriptionParams($plan['id'], $purchaseRequest, $purchase)
			], $accessToken, $errors);
		}
		else
		{
			$setupResponse = $this->makePayPalRequest('post', 'v2/checkout/orders', [
				'json' => $this->getPaymentParams($purchaseRequest, $purchase)
			], $accessToken, $errors);
		}

		if (isset($setupResponse['id']))
		{
			$approvalUrl = $purchase->recurring
				? $setupResponse['links'][0]['href'] // approve subscription
				: $setupResponse['links'][1]['href']; // authorize payment

			return $controller->redirect($approvalUrl);
		}
		else
		{
			throw $controller->exception($controller->error('Failed to create PayPal order: ' . reset($errors)));
		}
	}

	public function setupCallback(\XF\Http\Request $request)
	{
		$state = new CallbackState();

		$inputRaw = $request->getInputRaw();
		$state->webhookBody = $inputRaw;

		$input = @json_decode($inputRaw, true);
		$filtered = \XF::app()->inputFilterer()->filterArray($input ?: [], [
			'id' => 'str',
			'event_type' => 'str',
			'resource' => 'array',
		]);

		$state->webhookEventId = $filtered['id'];
		$state->eventType = $filtered['event_type'];
		$state->resource = $filtered['resource'];
		$state->webhookEvent = $input;

		if (!$state->resource)
		{
			return $state;
		}

		// ridiculous - why is the same thing described in at least three different ways??
		$state->requestKey = $state->resource['purchase_units'][0]['custom_id']
			?? $state->resource['disputed_transactions'][0]['custom']
			?? $state->resource['custom']
			?? $state->resource['custom_id']
			?? null;

		$state->transactionId = $state->resource['disputed_transactions'][0]['seller_transaction_id']
			?? $state->resource['id'];

		$state->subscriberId = $state->resource['billing_agreement_id'] ?? null;

		$paymentProfile = $state->getPaymentProfile();

		if (!$paymentProfile)
		{
			return $state;
		}

		$state->webhookSignature = [
			'auth_algo' => $request->getServer('HTTP_PAYPAL_AUTH_ALGO'),
			'cert_url' => $request->getServer('HTTP_PAYPAL_CERT_URL'),
			'transmission_id' => $request->getServer('HTTP_PAYPAL_TRANSMISSION_ID'),
			'transmission_sig' => $request->getServer('HTTP_PAYPAL_TRANSMISSION_SIG'),
			'transmission_time' => $request->getServer('HTTP_PAYPAL_TRANSMISSION_TIME'),
			'webhook_id' => $paymentProfile->options['webhook_id'],
			'webhook_event' => $state->webhookEvent,
		];

		return $state;
	}

	public function validateCallback(CallbackState $state)
	{
		if ($this->isEventSkippable($state))
		{
			// skip any irrelevant webhooks
			$state->httpCode = 200;
			return false;
		}

		if (!$this->verifyWebhookSignature($state))
		{
			$state->logType = 'error';
			$state->logMessage = 'Webhook received from PayPal could not be verified as being valid.';
			$state->httpCode = 400;

			return false;
		}

		return true;
	}

	protected function isEventSkippable(CallbackState $state): bool
	{
		$eventType = $state->eventType;

		if (!in_array($eventType, $this->getActionableEvents()))
		{
			return true;
		}

		return false;
	}

	protected function getActionableEvents(): array
	{
		return [
			'CHECKOUT.ORDER.APPROVED',
			'PAYMENT.CAPTURE.COMPLETED',
			'PAYMENT.CAPTURE.REFUNDED',
			'PAYMENT.SALE.COMPLETED',
			'PAYMENT.SALE.REFUNDED',
			'CUSTOMER.DISPUTE.CREATED',
			'CUSTOMER.DISPUTE.RESOLVED',
		];
	}

	protected function verifyWebhookSignature(CallbackState $state): bool
	{
		$options = $state->getPaymentProfile()->options;
		$accessToken = $this->getAccessToken($options['client_id'], $options['secret_key']);

		if (!$accessToken)
		{
			return false;
		}

		$response = $this->makePayPalRequest('post', 'v1/notifications/verify-webhook-signature', [
			'json' => $state->webhookSignature,
		], $accessToken, $errors);

		if ($errors)
		{
			return false;
		}

		return $response['verification_status'] === 'SUCCESS';
	}

	public function validateTransaction(CallbackState $state)
	{
		if (!$state->requestKey)
		{
			$state->logType = 'info';
			$state->logMessage = 'No purchase request key. Unrelated payment, no action to take.';
			return false;
		}

		if (!$state->getPurchaseRequest())
		{
			$state->logType = 'info';
			$state->logMessage = 'Invalid request key. Unrelated payment, no action to take.';
			return false;
		}

		if (!$state->transactionId && !$state->subscriberId)
		{
			$state->logType = 'info';
			$state->logMessage = 'No transaction or subscriber ID. No action to take.';
			return false;
		}

		return true;
	}

	public function validatePurchaseRequest(CallbackState $state)
	{
		// validated in validateTransaction
		return true;
	}

	public function validateCost(CallbackState $state)
	{
		if ($state->eventType === 'CUSTOMER.DISPUTE.CREATED' || $state->eventType === 'CUSTOMER.DISPUTE.RESOLVED')
		{
			// cost validation does not matter at this point
			return true;
		}

		$purchaseRequest = $state->getPurchaseRequest();

		$currency = $purchaseRequest->cost_currency;
		$cost = $purchaseRequest->cost_amount;

		$amount = $state->resource['amount'] ?? $state->resource['purchase_units'][0]['amount'];

		$costValidated = (
			($amount['total'] ?? $amount['value']) === $cost
			&& ($amount['currency'] ?? $amount['currency_code']) === strtoupper($currency)
		);

		if (!$costValidated)
		{
			$state->logType = 'error';
			$state->logMessage = 'Invalid cost amount';
			return false;
		}

		return true;
	}

	public function getPaymentResult(CallbackState $state)
	{
		switch ($state->eventType)
		{
			case 'CHECKOUT.ORDER.APPROVED':
				if ($state->resource['status'] === 'APPROVED')
				{
					$options = $state->getPaymentProfile()->options;
					$accessToken = $this->getAccessToken($options['client_id'], $options['secret_key']);

					if (!$accessToken)
					{
						return;
					}

					$orderId = $state->resource['id'];

					$captureResponse = $this->makePayPalRequest(
						'post', "v2/checkout/orders/$orderId/capture", [], $accessToken, $errors
					);

					if ($errors || $captureResponse['status'] !== 'COMPLETED')
					{
						$state->logType = 'error';
						$state->logMessage = reset($errors) ?: 'Order capture failed.';
						return;
					}
				}
				break;

			case 'PAYMENT.CAPTURE.COMPLETED':
			case 'PAYMENT.SALE.COMPLETED':
				$status = $state->resource['state'] ?? $state->resource['status'];
				if (strtoupper($status) === 'COMPLETED')
				{
					$state->paymentResult = CallbackState::PAYMENT_RECEIVED;
				}
				break;

			case 'PAYMENT.SALE.REFUNDED':
			case 'PAYMENT.CAPTURE.REFUNDED':
				$status = $state->resource['state'] ?? $state->resource['status'];
				if (strtoupper($status) === 'COMPLETED')
				{
					$state->paymentResult = CallbackState::PAYMENT_REVERSED;
				}
				break;

			case 'CUSTOMER.DISPUTE.CREATED':
				$state->paymentResult = CallbackState::PAYMENT_REVERSED;
				break;

			case 'CUSTOMER.DISPUTE.RESOLVED':
				$outcome = $state->resource['dispute_outcome']['outcome_code'] ?? null;

				switch ($outcome)
				{
					case 'RESOLVED_BUYER_FAVOUR':
					case 'RESOLVED_WITH_PAYOUT':
						// no action, already reversed
						break;

					case 'RESOLVED_SELLER_FAVOUR':
					case 'CANCELED_BY_BUYER':
						$state->paymentResult = CallbackState::PAYMENT_REINSTATED;
						break;
				}
				break;
		}
	}

	public function prepareLogData(CallbackState $state)
	{
		$state->logDetails = $state->webhookEvent;
	}
}