<?php

namespace XF\Pub\Controller;

use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;

use function count, in_array;

class FeaturedContent extends AbstractController
{
	public function actionIndex(ParameterBag $params): AbstractReply
	{
		return $this->rerouteController(__CLASS__, 'list', $params);
	}

	public function actionList(ParameterBag $params): AbstractReply
	{
		$featureRepo = $this->getFeatureRepo();

		$finder = $featureRepo->findFeaturedContent();
		$filters = $this->getFilterInput();
		$this->applyFilters($finder, $filters);

		$page = $this->filterPage($params->page);
		$perPage = $this->options()->discussionsPerPage; // TODO: custom option?
		$total = $finder->total();

		$this->assertValidPage($page, $perPage, $total, 'featured');
		$this->assertCanonicalUrl(
			$this->buildLink('featured', null, ['page' => $page])
		);

		/** @var \XF\Mvc\Entity\ArrayCollection $features */
		$features = $finder->limitByPage($page, $perPage)->fetch();
		$featureRepo->addContentToFeatures($features);
		$features = $features->filter(
			function(\XF\Entity\FeaturedContent $feature)
			{
				return $feature->canView() && !$feature->isIgnored();
			}
		);

		$displayFilters = $this->shouldDisplayFilters();

		$viewParams = [
			'features' => $features,

			'displayFilters' => $displayFilters,
			'filters' => $filters,

			'page' => $page,
			'perPage' => $perPage,
			'total' => $total
		];
		return $this->view(
			'XF:FeaturedContent\List',
			'featured_content_list',
			$viewParams
		);
	}

	protected function shouldDisplayFilters(): bool
	{
		$contentTypes = $this->getFeatureRepo()->getSupportedContentTypes();
		return count($contentTypes) > 1;
	}

	public function actionFilters(): AbstractReply
	{
		$filters = $this->getFilterInput();

		if ($this->filter('apply', 'bool'))
		{
			return $this->redirect($this->buildLink('featured', null, $filters));
		}

		$contentTypes = $this->getFeatureRepo()->getSupportedContentTypes();

		$viewParams = [
			'filters' => $filters,
			'contentTypes' => $contentTypes
		];
		return $this->view(
			'XF:FeaturedContent\Filters',
			'featured_content_filters',
			$viewParams
		);
	}

	/**
	 * @return mixed[]
	 */
	protected function getFilterInput(): array
	{
		$filters = [];

		$input = $this->filter([
			'content_type' => 'str'
		]);

		if (in_array(
			$input['content_type'],
			$this->getFeatureRepo()->getSupportedContentTypes()
		))
		{
			$filters['content_type'] = $input['content_type'];
		}

		return $filters;
	}

	/**
	 * @param mixed[] $filters
	 */
	protected function applyFilters(
		\XF\Mvc\Entity\Finder $finder,
		array $filters
	)
	{
		if (!empty($filters['content_type']))
		{
			$finder->where('content_type', $filters['content_type']);
		}
	}

	protected function getFeatureRepo(): \XF\Repository\FeaturedContent
	{
		return $this->repository('XF:FeaturedContent');
	}

	public static function getActivityDetails(array $activities): \XF\Phrase
	{
		return \XF::phrase(\XF::phrase('viewing_featured_content'));
	}
}
