<?php

namespace XF\Repository;

use XF\Mvc\Entity\Repository;
use XF\Service\Icon\SpriteGenerator;
use XF\Util\File;

use function array_key_exists, in_array;

class Icon extends Repository
{
	/**
	 * @var string
	 */
	public const ICON_VERSION = '5.15.3';

	/**
	 * @var string
	 */
	public const ICON_DIRECTORY = 'styles/fa/';

	/**
	 * @var string[]
	 */
	public const USAGE_TYPES = ['sprite', 'standalone'];

	/**
	 * @var string[]
	 */
	public const ICON_VARIANTS = [
		'fab' => 'brands',
		'fad' => 'duotone',
		'fal' => 'light',
		'far' => 'regular',
		'fas' => 'solid'
	];

	/**
	 * @var string
	 */
	public const ICON_DATA_REGEX = '<svg [^>]*viewBox="(?P<viewBox>[^"]+)"[^>]*>.*?(?:<defs>.*<\/defs>)?(?P<icon>(?:<path [^>]*\/>)+)<\/svg>';

	/**
	 * @var string
	 */
	public const ICON_CLASS_REGEX = '^fa-(?!-)(?P<name>[a-z0-9-]+)$';

	/**
	 * @var string
	 */
	public const ICON_CLASS_BLOCKLIST_REGEX = '^fa-(xs|sm|lg|\d+x|fw|ul|li|rotate-\d+|flip-(horizontal|vertical|both)|spin|pulse|border|pull-(left|right)|stack(-\dx)?|inverse)$';

	/**
	 * @var string
	 */
	public const ICON_LESS_REGEX = 'fa-var-((?P<variant>brands|duotone|light|regular|solid)-)?(?P<name>[a-z0-9-]+)';

	/**
	 * @var string
	 */
	public const ICON_PHRASE_REGEX = '\{icon(:(?P<variant>fa[bdlrs]))?::(?<name>[a-z0-9-]+)\}';

	public function getSpriteIconUrl(
		string $variant,
		string $name,
		bool $canonical = false
	): string
	{
		$path = $this->getSpriteIconPath(
			$variant,
			$name,
			$this->options()->iconSpriteLastUpdate
		);

		return $this->app()->applyLocalDataUrl($path, $canonical);
	}

	public function getSpriteIconPath(
		string $variant,
		string $name,
		string $version = null
	): string
	{
		if ($version === null)
		{
			$path = sprintf('icons/%s.svg#%s', $variant, $name);
		}
		else
		{
			$path = sprintf('icons/%s.svg?v=%s#%s', $variant, $version, $name);
		}

		return $path;
	}

	public function getStandaloneIconUrl(
		string $variant,
		string $name,
		bool $canonical = false
	)
	{
		$path = $this->getStandaloneIconPath(
			$variant,
			$name,
			static::ICON_VERSION
		);

		$pathType = $canonical ? 'canonical' : 'root-base';

		$pather = $this->app()->container('request.pather');
		return $pather($path, $pathType);
	}

	public function getStandaloneIconPath(
		string $variant,
		string $name,
		string $version = null
	): string
	{
		$path = static::ICON_DIRECTORY . $variant . '/' . $name . '.svg';

		if ($version !== null)
		{
			$path .= '?v=' . $version;
		}

		return $path;
	}

	/**
	 * @return string[]
	 */
	public function getIconData(string $variant, string $name): ?array
	{
		$path = File::canonicalizePath($this->getStandaloneIconPath($variant, $name));
		if (!file_exists($path))
		{
			return null;
		}

		$file = file_get_contents($path);
		if ($file === false)
		{
			return null;
		}

		$regex = static::ICON_DATA_REGEX;
		if (!preg_match("/{$regex}/i", $file, $matches))
		{
			return null;
		}

		return [
			'viewBox' => $matches['viewBox'],
			'icon' => $matches['icon']
		];
	}

	public function enqueueUsageAnalyzer(?string $contentType = null): void
	{
		$uniqueId = 'iconUsage';
		if ($contentType !== null)
		{
			$uniqueId .= '-' . $contentType;
		}

		$this->app()->jobManager()->enqueueUnique(
			$uniqueId,
			'XF:IconUsage',
			['content_type' => $contentType],
			false
		);
	}

	/**
	 * @param string[][][][] $icons
	 */
	public function recordUsage(array $icons): void
	{
		$inserts = [];

		foreach ($icons AS $contentType => $groupedIcons)
		{
			foreach ($groupedIcons AS $contentId => $icons)
			{
				foreach ($icons AS $icon)
				{
					$inserts[] = [
						'content_type' => $contentType,
						'content_id' => $contentId,
						'usage_type' => $icon['usage_type'],
						'icon_variant' => $icon['icon_variant'],
						'icon_name' => $icon['icon_name']
					];
				}
			}
		}

		if ($inserts)
		{
			$this->db()->insertBulk('xf_icon_usage', $inserts);
		}
	}

	/**
	 * @return string[][]
	 */
	public function fetchUsageByType(string $usageType): array
	{
		if (!in_array($usageType, static::USAGE_TYPES))
		{
			throw new \InvalidArgumentException(
				"Invalid usage type: {$usageType}"
			);
		}

		$icons = $this->db()->fetchAll(
			'SELECT DISTINCT icon_variant, icon_name
				FROM xf_icon_usage
				WHERE usage_type = ?
				ORDER BY icon_name',
			[$usageType]
		);

		$groupedIcons = [];

		foreach ($icons AS $icon)
		{
			$groupedIcons[$icon['icon_variant']][] = $icon['icon_name'];
		}

		return $groupedIcons;
	}

	public function purgeUsageRecords(?string $contentType = null): void
	{
		if ($contentType === null)
		{
			$this->db()->emptyTable('xf_icon_usage');
		}
		else
		{
			$this->db()->delete(
				'xf_icon_usage',
				'content_type = ?',
				[$contentType]
			);
		}
	}

	public function runSpriteGenerator(): void
	{
		/** @var SpriteGenerator $spriteGenerator */
		$spriteGenerator = $this->app()->service(
			'XF:Icon\SpriteGenerator',
			$this->fetchUsageByType('sprite')
		);
		$spriteGenerator->generate();
	}

	/**
	 * @return string[]
	 */
	public function getDefaultVariants(): array
	{
		$weights = [];

		/** @var int[] $styleIds */
		$styleIds = array_merge(
			[0],
			$this->db()->fetchAllColumn('SELECT style_id FROM xf_style')
		);
		foreach ($styleIds AS $styleId)
		{
			if ($styleId === 0)
			{
				$style = $this->app()->container('style.fallback');
			}
			else
			{
				$style = $this->app()->style($styleId);
			}

			$weight = $style->getProperty('fontAwesomeWeight', 400);
			$weights[$styleId] = $weight;
		}

		return array_map(
			function (int $weight): string
			{
				return $this->getIconVariantFromWeight($weight);
			},
			$weights
		);
	}

	public function getIconVariantFromWeight(int $weight): string
	{
		switch ($weight)
		{
			case 300: return 'light';
			case 400: return 'regular';
			case 900: return 'solid';
		}

		return 'regular';
	}

	/**
	 * @return list<{name: string, variant: string}>
	 */
	public function getIconsFromClasses(string $classes): array
	{
		$names = [];
		$variants = [];

		$classes = explode(' ', $classes);
		$classRegex = static::ICON_CLASS_REGEX;
		$blocklistRegex = static::ICON_CLASS_BLOCKLIST_REGEX;
		$validVariants = static::ICON_VARIANTS;
		foreach ($classes AS $class)
		{
			if (preg_match("/{$blocklistRegex}/i", $class))
			{
				continue;
			}

			if (array_key_exists($class, $validVariants))
			{
				$variants[] = $validVariants[$class];
				continue;
			}

			if (preg_match("/{$classRegex}/i", $class, $matches))
			{
				$names[] = $matches['name'];
				continue;
			}
		}

		$icons = [];

		if (!$names)
		{
			return [];
		}

		if (!$variants)
		{
			$variants = [null];
		}

		foreach ($variants AS $variant)
		{
			foreach ($names AS $name)
			{
				$icons[] = [
					'name' => $name,
					'variant' => $variant
				];
			}
		}

		return $icons;
	}

	/**
	 * @return string[]
	 */
	public function getFontAwesome4IconMap(): array
	{
		return [
			'glass' => 'glass-martini',
			'star-o' => 'star',
			'remove' => 'times',
			'close' => 'times',
			'gear' => 'cog',
			'trash-o' => 'trash-alt',
			'file-o' => 'file',
			'clock-o' => 'clock',
			'arrow-circle-o-down' => 'arrow-alt-circle-down',
			'arrow-circle-o-up' => 'arrow-alt-circle-up',
			'play-circle-o' => 'play-circle',
			'repeat' => 'redo',
			'rotate-right' => 'redo',
			'refresh' => 'sync',
			'dedent' => 'outdent',
			'video-camera' => 'video',
			'picture-o' => 'image',
			'photo' => 'image',
			'pencil' => 'pencil-alt',
			'map-marker' => 'map-marker-alt',
			'pencil-square-o' => 'edit',
			'share-square-o' => 'share-square',
			'check-square-o' => 'check-square',
			'arrows' => 'arrows-alt',
			'times-circle-o' => 'times-circle',
			'check-circle-o' => 'check-circle',
			'mail-forward' => 'share',
			'warning' => 'exclamation-triangle',
			'calendar' => 'calendar-alt',
			'arrows-v' => 'arrows-alt-v',
			'arrows-h' => 'arrows-alt-h',
			'bar-chart' => 'chart-bar',
			'bar-chart-o' => 'chart-bar',
			'gears' => 'cogs',
			'thumbs-o-up' => 'thumbs-up',
			'thumbs-o-down' => 'thumbs-down',
			'heart-o' => 'heart',
			'sign-out' => 'sign-out-alt',
			'linkedin-square' => 'linkedin',
			'thumb-tack' => 'thumbtack',
			'external-link' => 'external-link-alt',
			'sign-in' => 'sign-in-alt',
			'lemon-o' => 'lemon',
			'square-o' => 'square',
			'bookmark-o' => 'bookmark',
			'facebook' => 'facebook-f',
			'feed' => 'rss',
			'hdd-o' => 'hdd',
			'hand-o-right' => 'hand-point-right',
			'hand-o-left' => 'hand-point-left',
			'hand-o-up' => 'hand-point-up',
			'hand-o-down' => 'hand-point-down',
			'arrows-alt' => 'expand-arrows-alt',
			'group' => 'users',
			'chain' => 'link',
			'scissors' => 'cut',
			'files-o' => 'copy',
			'floppy-o' => 'save',
			'navicon' => 'bars',
			'reorder' => 'bars',
			'google-plus' => 'google-plus-g',
			'money' => 'money-bill-alt',
			'unsorted' => 'sort',
			'sort-desc' => 'sort-down',
			'sort-asc' => 'sort-up',
			'linkedin' => 'linkedin-in',
			'rotate-left' => 'undo',
			'legal' => 'gavel',
			'tachometer' => 'tachometer-alt',
			'dashboard' => 'tachometer-alt',
			'comment-o' => 'comment',
			'comments-o' => 'comments',
			'flash' => 'bolt',
			'paste' => 'clipboard',
			'lightbulb-o' => 'lightbulb',
			'exchange' => 'exchange-alt',
			'cloud-download' => 'cloud-download-alt',
			'cloud-upload' => 'cloud-upload-alt',
			'bell-o' => 'bell',
			'cutlery' => 'utensils',
			'file-text-o' => 'file-alt',
			'building-o' => 'building',
			'hospital-o' => 'hospital',
			'tablet' => 'tablet-alt',
			'mobile' => 'mobile-alt',
			'mobile-phone' => 'mobile-alt',
			'circle-o' => 'circle',
			'mail-reply' => 'reply',
			'folder-o' => 'folder',
			'folder-open-o' => 'folder-open',
			'smile-o' => 'smile',
			'frown-o' => 'frown',
			'meh-o' => 'meh',
			'keyboard-o' => 'keyboard',
			'flag-o' => 'flag',
			'mail-reply-all' => 'reply-all',
			'star-half-o' => 'star-half',
			'star-half-empty' => 'star-half',
			'star-half-full' => 'star-half',
			'code-fork' => 'code-branch',
			'chain-broken' => 'unlink',
			'shield' => 'shield-alt',
			'calendar-o' => 'calendar',
			'ticket' => 'ticket-alt',
			'minus-square-o' => 'minus-square',
			'level-up' => 'level-up-alt',
			'level-down' => 'level-down-alt',
			'pencil-square' => 'pen-square',
			'external-link-square' => 'external-link-square-alt',
			'caret-square-o-down' => 'caret-square-down',
			'toggle-down' => 'caret-square-down',
			'caret-square-o-up' => 'caret-square-up',
			'toggle-up' => 'caret-square-up',
			'caret-square-o-right' => 'caret-square-right',
			'toggle-right' => 'caret-square-right',
			'eur' => 'euro-sign',
			'euro' => 'euro-sign',
			'gbp' => 'pound-sign',
			'usd' => 'dollar-sign',
			'dollar' => 'dollar-sign',
			'inr' => 'rupee-sign',
			'rupee' => 'rupee-sign',
			'jpy' => 'yen-sign',
			'cny' => 'yen-sign',
			'rmb' => 'yen-sign',
			'yen' => 'yen-sign',
			'rub' => 'ruble-sign',
			'ruble' => 'ruble-sign',
			'rouble' => 'ruble-sign',
			'krw' => 'won-sign',
			'won' => 'won-sign',
			'bitcoin' => 'btc',
			'file-text' => 'file-alt',
			'sort-alpha-asc' => 'sort-alpha-down',
			'sort-alpha-desc' => 'sort-alpha-up',
			'sort-amount-asc' => 'sort-amount-down',
			'sort-amount-desc' => 'sort-amount-up',
			'sort-numeric-asc' => 'sort-numeric-down',
			'sort-numeric-desc' => 'sort-numeric-up',
			'youtube-play' => 'youtube',
			'bitbucket-square' => 'bitbucket',
			'long-arrow-down' => 'long-arrow-alt-down',
			'long-arrow-up' => 'long-arrow-alt-up',
			'long-arrow-left' => 'long-arrow-alt-left',
			'long-arrow-right' => 'long-arrow-alt-right',
			'gittip' => 'gratipay',
			'sun-o' => 'sun',
			'moon-o' => 'moon',
			'arrow-circle-o-right' => 'arrow-alt-circle-right',
			'arrow-circle-o-left' => 'arrow-alt-circle-left',
			'caret-square-o-left' => 'caret-square-left',
			'toggle-left' => 'caret-square-left',
			'dot-circle-o' => 'dot-circle',
			'try' => 'lira-sign',
			'turkish-lira' => 'lira-sign',
			'plus-square-o' => 'plus-square',
			'institution' => 'university',
			'bank' => 'university',
			'mortar-board' => 'graduation-cap',
			'spoon' => 'utensil-spoon',
			'automobile' => 'car',
			'cab' => 'taxi',
			'envelope-o' => 'envelope',
			'file-pdf-o' => 'file-pdf',
			'file-word-o' => 'file-word',
			'file-excel-o' => 'file-excel',
			'file-powerpoint-o' => 'file-powerpoint',
			'file-image-o' => 'file-image',
			'file-photo-o' => 'file-image',
			'file-picture-o' => 'file-image',
			'file-archive-o' => 'file-archive',
			'file-zip-o' => 'file-archive',
			'file-audio-o' => 'file-audio',
			'file-sound-o' => 'file-audio',
			'file-video-o' => 'file-video',
			'file-movie-o' => 'file-video',
			'file-code-o' => 'file-code',
			'life-bouy' => 'life-ring',
			'life-buoy' => 'life-ring',
			'life-saver' => 'life-ring',
			'support' => 'life-ring',
			'circle-o-notch' => 'circle-notch',
			'ra' => 'rebel',
			'resistance' => 'rebel',
			'ge' => 'empire',
			'y-combinator-square' => 'hacker-news',
			'yc-square' => 'hacker-news',
			'wechat' => 'weixin',
			'send' => 'paper-plane',
			'paper-plane-o' => 'paper-plane',
			'send-o' => 'paper-plane',
			'circle-thin' => 'circle',
			'header' => 'heading',
			'sliders' => 'sliders-h',
			'futbol-o' => 'futbol',
			'soccer-ball-o' => 'futbol',
			'newspaper-o' => 'newspaper',
			'bell-slash-o' => 'bell-slash',
			'trash' => 'trash-alt',
			'eyedropper' => 'eye-dropper',
			'area-chart' => 'chart-area',
			'pie-chart' => 'chart-pie',
			'line-chart' => 'chart-line',
			'cc' => 'closed-captioning',
			'ils' => 'shekel-sign',
			'shekel' => 'shekel-sign',
			'sheqel' => 'shekel-sign',
			'meanpath' => 'font-awesome',
			'diamond' => 'gem',
			'intersex' => 'transgender',
			'facebook-official' => 'facebook',
			'hotel' => 'bed',
			'yc' => 'y-combinator',
			'battery-4' => 'battery-full',
			'battery' => 'battery-full',
			'battery-3' => 'battery-three-quarters',
			'battery-2' => 'battery-half',
			'battery-1' => 'battery-quarter',
			'battery-0' => 'battery-empty',
			'sticky-note-o' => 'sticky-note',
			'hourglass-o' => 'hourglass',
			'hourglass-1' => 'hourglass-start',
			'hourglass-2' => 'hourglass-half',
			'hourglass-3' => 'hourglass-end',
			'hand-rock-o' => 'hand-rock',
			'hand-grab-o' => 'hand-rock',
			'hand-paper-o' => 'hand-paper',
			'hand-stop-o' => 'hand-paper',
			'hand-scissors-o' => 'hand-scissors',
			'hand-lizard-o' => 'hand-lizard',
			'hand-spock-o' => 'hand-spock',
			'hand-pointer-o' => 'hand-pointer',
			'hand-peace-o' => 'hand-peace',
			'television' => 'tv',
			'calendar-plus-o' => 'calendar-plus',
			'calendar-minus-o' => 'calendar-minus',
			'calendar-times-o' => 'calendar-times',
			'calendar-check-o' => 'calendar-check',
			'map-o' => 'map',
			'commenting' => 'comment-dots',
			'commenting-o' => 'comment-dots',
			'vimeo' => 'vimeo-v',
			'credit-card-alt' => 'credit-card',
			'pause-circle-o' => 'pause-circle',
			'stop-circle-o' => 'stop-circle',
			'wheelchair-alt' => 'accessible-icon',
			'question-circle-o' => 'question-circle',
			'volume-control-phone' => 'phone-volume',
			'asl-interpreting' => 'american-sign-language-interpreting',
			'deafness' => 'deaf',
			'hard-of-hearing' => 'deaf',
			'signing' => 'sign-language',
			'google-plus-official' => 'google-plus',
			'google-plus-circle' => 'google-plus',
			'fa' => 'font-awesome',
			'handshake-o' => 'handshake',
			'envelope-open-o' => 'envelope-open',
			'address-book-o' => 'address-book',
			'vcard' => 'address-card',
			'address-card-o' => 'address-card',
			'vcard-o' => 'address-card',
			'user-circle-o' => 'user-circle',
			'user-o' => 'user',
			'drivers-license' => 'id-card',
			'id-card-o' => 'id-card',
			'drivers-license-o' => 'id-card',
			'thermometer-4' => 'thermometer-full',
			'thermometer' => 'thermometer-full',
			'thermometer-3' => 'thermometer-three-quarters',
			'thermometer-2' => 'thermometer-half',
			'thermometer-1' => 'thermometer-quarter',
			'thermometer-0' => 'thermometer-empty',
			'bathtub' => 'bath',
			's15' => 'bath',
			'times-rectangle' => 'window-close',
			'window-close-o' => 'window-close',
			'times-rectangle-o' => 'window-close',
			'eercast' => 'sellcast',
			'snowflake-o' => 'snowflake',
		];
	}

	/**
	 * @return string[]
	 */
	public function getButtonIconMap(): array
	{
		return [
			'add' => 'fa-plus-square',
			'confirm' => 'fa-check',
			'write' => 'fa-edit',
			'import' => 'fa-upload',
			'export' => 'fa-download',
			'download' => 'fa-download',
			'redirect' => 'fa-external-link',
			'disable' => 'fa-power-off',
			'edit' => 'fa-edit',
			'save' => 'fa-save',
			'reply' => 'fa-reply',
			'quote' => 'fa-quote-left',
			'purchase' => 'fa-credit-card',
			'payment' => 'fa-credit-card',
			'convert' => 'fa-bolt',
			'search' => 'fa-search',
			'sort' => 'fa-sort',
			'upload' => 'fa-upload',
			'attach' => 'fa-paperclip',
			'login' => 'fa-lock',
			'rate' => 'fa-star',
			'config' => 'fa-cog',
			'refresh' => 'fa-sync-alt',
			'translate' => 'fa-globe',
			'vote' => 'fa-check-circle',
			'result' => 'fa-chart-bar',
			'history' => 'fa-history',
			'cancel' => 'fa-ban',
			'close' => 'fa-times',
			'preview' => 'fa-eye',
			'conversation' => 'fa-comments',
			'bolt' => 'fa-bolt',
			'list' => 'fa-list',
			'prev' => 'fa-chevron-left',
			'next' => 'fa-chevron-right',
			'markRead' => 'fa-check-square',
			'user' => 'fa-user',
			'userCircle' => 'fa-user-circle',

			'notificationsOn' => 'fa-bell',
			'notificationsOff' => 'fa-bell-slash',

			'show' => 'fa-eye',
			'hide' => 'fa-eye-slash',

			'merge' => 'fa-compress',
			'move' => 'fa-share',
			'copy' => 'fa-copy',
			'approve' => 'fa-shield',
			'unapprove' => 'fa-shield',
			'delete' => 'fa-trash-alt',
			'undelete' => 'fa-trash-alt',
			'stick' => 'fa-thumbtack',
			'unstick' => 'fa-thumbtack',
			'lock' => 'fa-lock',
			'unlock' => 'fa-unlock'
		];
	}

	/**
	 * @return string[]
	 */
	public function getExtraIcons(): array
	{
		$icons = [];

		$icons[] = 'fa-exclamation-triangle'; // user activity error icon

		// style variation icons
		$icons[] = 'fa-adjust';
		$icons[] = 'fa-sun';
		$icons[] = 'fa-moon';

		$this->app()->fire('icon_usage_analyzer_extra', [&$icons]);

		return $icons;
	}
}
