<?php

namespace XF\Repository;

use XF\Entity\OAuthCode;
use XF\Mvc\Entity\Repository;

class OAuth extends Repository
{
	public const CLIENT_TYPE_PUBLIC = 'public';
	public const CLIENT_TYPE_CONFIDENTIAL = 'confidential';

	public const RESPONSE_TYPE_CODE = 'code';

	public const CODE_CHALLENGE_METHOD_S256 = 'S256';

	public function findClientsForList(): \XF\Mvc\Entity\Finder
	{
		return $this->finder(\XF\Finder\OAuthClient::class)
			->setDefaultOrder('creation_date', 'desc');
	}

	public function findActiveTokensForUser(\XF\Entity\User $user): \XF\Mvc\Entity\Finder
	{
		return $this->finder(\XF\Finder\OAuthToken::class)
			->with('OAuthClient')
			->where('user_id', $user->user_id)
			->order('issue_date', 'desc');
	}

	public function getConnectedClientsForUser(\XF\Entity\User $user = null): \XF\Mvc\Entity\AbstractCollection
	{
		if (!$user)
		{
			$user = \XF::visitor();
		}

		$uniqueClientIds = $this->db()->fetchAllColumn("
			SELECT DISTINCT client_id
			FROM xf_oauth_token
			WHERE user_id = ?
			AND revoked_date = 0
			ORDER BY issue_date DESC
		", $user->user_id);

		return $this->finder(\XF\Finder\OAuthClient::class)
			->where('client_id', $uniqueClientIds)
			->where('active', 1)
			->order('client_id', 'desc')
			->fetch();
	}

	public function getScopesForTokens(\XF\Entity\OAuthClient $client)
	{
		$scopesFromTokens = $this->finder(\XF\Entity\OAuthToken::class)
			->where('client_id', $client->client_id)
			->where('user_id', \XF::visitor()->user_id)
			->fetch()
			->pluckNamed('scopes', 'token_id');

		$scopes = array_merge(...$scopesFromTokens);
		$scopes = array_keys($scopes);

		$apiScopes = $this->finder(\XF\Entity\ApiScope::class)
			->whereIds($scopes)
			->fetch();

		return $apiScopes;
	}

	public function revokeClientForUser(\XF\Entity\OAuthClient $client, \XF\Entity\User $user = null): void
	{
		if (!$user)
		{
			$user = \XF::visitor();
		}

		$this->db()->update('xf_oauth_token', ['revoked_date' => \XF::$time], 'client_id = ? AND user_id = ?', [$client->client_id, $user->user_id]);
	}

	public function pruneExpiredCodes(?int $cutOff = null): int
	{
		if ($cutOff === null)
		{
			$cutOff = \XF::$time - 86400 * 14;
		}

		return $this->db()->delete('xf_oauth_code', 'expiry_date < ?', $cutOff);
	}

	public function pruneAuthRequests(?int $cutOff = null): int
	{
		if ($cutOff === null)
		{
			$cutOff = \XF::$time - 86400 * 30;
		}

		return $this->db()->delete('xf_oauth_request', 'request_date < ?', $cutOff);
	}
}