<?php

namespace XF\Search\Data;

use XF\Mvc\Entity\Entity;
use XF\Search\IndexRecord;
use XF\Search\MetadataStructure;
use XF\Search\Query\MetadataConstraint;
use XF\Search\Query\SqlOrder;
use XF\Search\Query\TableReference;
use XF\Util\Arr;

class ConversationMessage extends AbstractData
{
	public function getIndexData(Entity $entity): ?IndexRecord
	{
		/** @var \XF\Entity\ConversationMessage $entity */
		if (!$entity->Conversation)
		{
			return null;
		}

		if ($entity->isFirstMessage())
		{
			$conversation = $entity->Conversation;
			$handler = $this->searcher->handler('conversation');
			return $handler->getIndexData($conversation);
		}

		return IndexRecord::create('conversation_message', $entity->message_id, [
			'message' => $entity->message_,
			'date' => $entity->message_date,
			'user_id' => $entity->user_id,
			'discussion_id' => $entity->conversation_id,
			'metadata' => $this->getMetaData($entity)
		]);
	}

	/**
	 * @return array<string, mixed>
	 */
	protected function getMetaData(\XF\Entity\ConversationMessage $entity): array
	{
		$conversation = $entity->Conversation;
		$recipients = array_keys($conversation->recipients);
		$recipients[]  = $conversation->user_id;

		return [
			'conversation' => $entity->conversation_id,
			'recipients'   => $recipients,
		];
	}

	public function setupMetadataStructure(MetadataStructure $structure): void
	{
		$structure->addField('conversation', MetadataStructure::INT);
		$structure->addField('recipients', MetadataStructure::INT);
	}

	public function getResultDate(Entity $entity): int
	{
		/** @var \XF\Entity\ConversationMessage $entity */
		return $entity->message_date;
	}

	/**
	 * @return array<string, mixed>
	 */
	public function getTemplateData(Entity $entity, array $options = []): array
	{
		return [
			'message' => $entity,
			'options' => $options,
		];
	}

	/**
	 * @param bool $forView
	 *
	 * @return list<string>
	 */
	public function getEntityWith($forView = false): array
	{
		$with = ['Conversation'];

		if ($forView)
		{
			$with[] =  'User';
			$with[] = 'Conversation.Users|' . \XF::visitor()->user_id;
		}

		return $with;
	}

	/**
	 * @return list<string>
	 */
	public function getSearchableContentTypes(): array
	{
		return ['conversation', 'conversation_message'];
	}

	/**
	 * @return array<string, mixed>
	 */
	public function getSearchFormTab(): ?array
	{
		if (!\XF::visitor()->user_id)
		{
			return null;
		}

		return [
			'title' => \XF::phrase('search_direct_messages'),
			'order' => 1100,
		];
	}

	/**
	 * @param array<string, mixed> $urlConstraints
	 */
	public function applyTypeConstraintsFromInput(
		\XF\Search\Query\Query $query,
		\XF\Http\Request $request,
		array &$urlConstraints
	): void
	{
		$recipients = $request->filter('c.recipients', 'str');
		if ($recipients)
		{
			$recipients = Arr::stringToArray($recipients, '/,\s*/');
			if ($recipients)
			{
				/** @var \XF\Repository\User $userRepo */
				$userRepo = \XF::repository('XF:User');
				$matchedUsers = $userRepo->getUsersByNames($recipients, $notFound);
				if ($notFound)
				{
					$query->error(
						'users',
						\XF::phrase('following_members_not_found_x', [
							'members' => implode(', ', $notFound)
						])
					);
				}
				else
				{
					$query->withMetadata('recipients', $matchedUsers->keys());
					$urlConstraints['recipients'] = implode(', ', $recipients);
				}
			}
		}

		$minReplyCount = $request->filter('c.min_reply_count', 'uint');
		if ($minReplyCount)
		{
			$query->withSql(new \XF\Search\Query\SqlConstraint(
				'conversation.reply_count >= %s',
				$minReplyCount,
				$this->getConversationQueryTableReference()
			));
		}
		else
		{
			unset($urlConstraints['min_reply_count']);
		}

		$conversationId = $request->filter('c.conversation', 'uint');
		if ($conversationId)
		{
			$query->withMetadata('conversation', $conversationId);

			if ($query instanceof \XF\Search\Query\KeywordQuery)
			{
				$query->inTitleOnly(false);
			}
		}
		else
		{
			unset($urlConstraints['conversation']);
		}
	}

	public function getGroupByType(): string
	{
		return 'conversation';
	}

	/**
	 * @param string|null $order
	 */
	public function getTypeOrder($order): ?SqlOrder
	{
		if ($order === 'replies')
		{
			return new \XF\Search\Query\SqlOrder(
				'conversation.reply_count DESC',
				$this->getConversationQueryTableReference()
			);
		}

		return null;
	}

	protected function getConversationQueryTableReference(): TableReference
	{
		return new TableReference(
			'conversation',
			'xf_conversation_master',
			'conversation.conversation_id = search_index.discussion_id'
		);
	}

	/**
	 * @param bool $isOnlyType
	 *
	 * @return list<MetadataConstraint>
	 */
	public function getTypePermissionConstraints(
		\XF\Search\Query\Query $query,
		$isOnlyType
	): array
	{
		if (!$isOnlyType)
		{
			return [];
		}

		return [
			new MetadataConstraint('recipients', \XF::visitor()->user_id)
		];
	}

	/**
	 * @param list<int> $resultIds
	 */
	public function canIncludeInResults(Entity $entity, array $resultIds): bool
	{
		/** @var \XF\Entity\ConversationMessage $entity */
		if (
			isset($resultIds['conversation-' . $entity->conversation_id]) &&
			$entity->isFirstMessage()
		)
		{
			return false;
		}

		return true;
	}
}