<?php

namespace XF\Widget;

class FeaturedContent extends AbstractWidget
{
	/**
	 * @var mixed[]
	 */
	protected $defaultOptions = [
		'contextual' => false,
		'contextual_hidden' => false,
		'limit' => 5,
		'style' => 'simple',
		'snippet_length' => 500,
		'content_type' => null,
		'content_container_id' => null
	];

	/**
	 * @param string $context
	 *
	 * @return mixed[]
	 */
	protected function getDefaultTemplateParams($context)
	{
		$params = parent::getDefaultTemplateParams($context);

		if ($context == 'options')
		{
			$params['contentTypes'] = $this->getFeatureRepo()->getSupportedContentTypes();
		}

		return $params;
	}

	/**
	 * @return \XF\Widget\WidgetRenderer|null
	 */
	public function render()
	{
		$options = $this->options;
		if ($options['contextual'])
		{
			$contextualOptions = $this->getContextualOptions();
			if ($contextualOptions)
			{
				$options = array_replace($options, $contextualOptions);
			}
			else if ($options['contextual_hidden'])
			{
				return null;
			}
		}

		$limit = $options['limit'];
		$style = $options['style'];
		$snippetLength = $options['snippet_length'];
		$contentType = $options['content_type'];
		$containerId = $options['content_container_id'];

		$featureRepo = $this->getFeatureRepo();
		$finder = $featureRepo->findFeaturedContent();

		if ($contentType)
		{
			$finder->where('content_type', $contentType);
		}

		if ($containerId)
		{
			$finder->where('content_container_id', $containerId);
		}

		/** @var \XF\Mvc\Entity\ArrayCollection $features */
		$features = $finder->fetch(max($limit * 2, 10));
		$featureRepo->addContentToFeatures($features);

		$features = $features->filter(
			function(\XF\Entity\FeaturedContent $feature)
			{
				return $feature->canView() && !$feature->isIgnored();
			}
		);

		$total = $features->count();
		$features = $features->slice(0, $limit, true);
		$hasMore = $total > $features->count();

		$title = $this->getTitle();

		$router = $this->app->router('public');

		$linkParams = [];
		if ($contentType)
		{
			$linkParams['content_type'] = $contentType;
		}

		$link = $router->buildLink('featured', null, $linkParams);

		$viewParams = [
			'title' => $title,
			'link' => $link,
			'features' => $features,

			'style' => $style,
			'snippetLength' => $snippetLength,
			'hasMore' => $hasMore
		];
		return $this->renderer('widget_featured_content', $viewParams);
	}

	protected function getContextualOptions(): array
	{
		$forum = $this->contextParams['forum'] ?? null;
		if ($forum && $forum instanceof \XF\Entity\Forum)
		{
			return [
				'content_type' => 'thread',
				'content_container_id' => $forum->node_id
			];
		}

		return [];
	}

	/**
	 * @param \XF\Phrase|null $error
	 */
	public function verifyOptions(
		\XF\Http\Request $request,
		array &$options,
		&$error = null
	): bool
	{
		$options = $request->filter([
			'contextual' => 'bool',
			'contextual_hidden' => 'bool',
			'limit' => 'uint',
			'style' => 'str',
			'snippet_length' => 'uint',
			'content_type' => 'str'
		]);

		if ($options['limit'] < 1)
		{
			$options['limit'] = 1;
		}

		return true;
	}

	protected function getFeatureRepo(): \XF\Repository\FeaturedContent
	{
		return $this->repository('XF:FeaturedContent');
	}
}