<?php

namespace Doctrine\Common\Cache;

use Psr\Cache\CacheItemInterface;
use Symfony\Component\Cache\Adapter\AdapterInterface;

/**
 * An adapter for providing Symfony cache adapters through the Doctrine cache
 * interface.
 */
class CacheProvider
{
	/**
	 * @var AdapterInterface
	 */
	protected $adapter;

	public function __construct(AdapterInterface $adapter)
	{
		$this->adapter = $adapter;
	}

	public function getAdapter(): AdapterInterface
	{
		return $this->adapter;
	}

	/**
	 * @return mixed
	 */
	public function fetch(string $id)
	{
		$item = $this->adapter->getItem($id);
		if (!$item->isHit())
		{
			return false;
		}

		return $item->get();
	}

	/**
	 * @param list<string> $ids
	 *
	 * @return array<string, mixed>
	 */
	public function fetchMultiple(array $ids): array
	{
		$items = $this->adapter->getItems($ids);

		$results = [];

		foreach ($items AS $id => $item)
		{
			if (!$item->isHit())
			{
				continue;
			}

			$results[$id] = $item->get();
		}

		return $results;
	}

	public function contains(string $id): bool
	{
		return $this->adapter->hasItem($id);
	}

	/**
	 * @param mixed $data
	 */
	public function save(string $id, $data, int $lifeTime = 0): bool
	{
		$item = $this->getCacheItemForSave($id, $data, $lifeTime);
		return $this->adapter->save($item);
	}

	/**
	 * @param array<string, mixed> $keysAndValues
	 */
	public function saveMultiple(array $keysAndValues, int $lifeTime = 0): bool
	{
		foreach ($keysAndValues AS $id => $data)
		{
			$item = $this->getCacheItemForSave($id, $data, $lifeTime);
			$this->adapter->saveDeferred($item);
		}

		return $this->adapter->commit();
	}

	/**
	 * @param mixed $data
	 */
	protected function getCacheItemForSave(
		string $id,
		$data,
		int $lifeTime = 0
	): CacheItemInterface
	{
		$item = $this->adapter->getItem($id);

		$item->set($data);

		if ($lifeTime !== 0)
		{
			$item->expiresAfter($lifeTime);
		}

		return $item;
	}

	public function delete(string $id): bool
	{
		return $this->adapter->deleteItem($id);
	}

	/**
	 * @param list<string> $ids
	 */
	public function deleteMultiple(array $ids): bool
	{
		return $this->adapter->deleteItems($ids);
	}

	public function flushAll(): bool
	{
		return $this->adapter->clear();
	}


	public function deleteAll(): bool
	{
		return $this->adapter->clear();
	}

	/**
	 * @deprecated
	 */
	public function getStats(): array
	{
		trigger_error(
			'The cache getStats method has been deprecated',
			E_USER_DEPRECATED
		);

		return [];
	}
}
